/*****************************************************************************
 * F1 Racing diteur/simulateur
 * Fichier: widgettrack.h
 *
 * (c) Leroux philippe, 2003
 *
 *
 * This program is distributed under the terms of the GNU General Public
 *
 * License:
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****************************************************************************/
#ifndef WIDGETTRACK_H
#define WIDGETTRACK_H

//============================================================================
//	WidgetTrack:
//		Gestion de l'affichage d'une piste.
//			- chargement d'une piste.
//			- modification des pixels d'une piste.
//			- enregistrement d'une piste.
//			- test la validit des bordures de la piste.
//
//============================================================================

#include <qwidget.h>
#include <qvector.h>

#include "pixel/pixel.h"
#include "dimensions.h"

#include <cassert>

class TrackPixmap;
class WidgetTrack : public QWidget
{
Q_OBJECT
public:
	WidgetTrack( int widthPen, QWidget* parent, const char* name );

	enum State { Empty, Changed, Saved };
	State			state() const { return state_; }

	// Charge la track du fichier 'name'
	bool			load( const QString& name, bool clean=true );
	// Enregiste la track courante dans le fichier 'name'.
	bool			save( const QString& name );


	const Pixel&	pixel( int num ) const;

	//	shadowPixel, join, select et cleanSelect ne modifient pas la piste.
	//	join				: Trace une ligne (efface le trac prcdent)
	//	select				: Trace un rectangle.
	//	cleanSelect			: efface la ligne trace par select ou join.
	//	cancelCleanSelect	: empcher que le forme dessine par join ou select
	//						  soit efface, sauf si la piste est redessine.
	void			shadowPixel( const QColor&, int numPixel );
	void			join( const QColor&, const Pixel& ori, const Pixel& dest );
	void			select( const QColor&,const Pixel& ori,const Pixel& dest );
	void			select( const QColor&, int numPixel0, int numPixel1 );
	void			cleanSelect( bool erase=true );
	void			cancelCleanSelect();

	//	Colorise avec la couleur les pixels situs dans la zone
	//	dlimite pour les deux pixels passs en paramtre.
	TrackPixmap*	fill( const QColor&, Pixel*, Pixel* );
	TrackPixmap*	fill( const QColor&, int numPixel0, int numPixel1 );

	//	Dessine l'image 'imageBefore', si 'redo' est non null renvoie
	//	l'ancienne image de la zone de imageBefore.
	void 			play( const TrackPixmap& imageBefore,
													TrackPixmap* redo=NULL );

	//	Les 3 fonctions ne modifie pas la piste : ie state_ n'est pas affect.
	//	moveCar renvoient le n de pixel invalide, -1 si ok.
	int				moveCar( const QColor& c, const QColor& t,
											const Pixel& of, const Pixel& to );
	void			paintPixel( const QColor& c, Pixel& pixel );
	void			paintPixel( const QColor& c, int num );

	// Vrifie et corrige les bordures de la track.
	bool			normalizeTrackBorder();
	bool			checkStarterArea();

	//	Renvoie true si 'c' est utilis dans la piste.
	bool			colorUsed( const QColor& c ) const;

	void			setWidthPen( int w )	{ widthPen_ = w; }

	override QSize	minimumSizeHint() const { return minimumSize_; }

signals:
	void warning( bool error, const QString& );

	//	Sur la slection du boutton de la sourie, envoie le pixel sous
	//	le curseur.
	void mouseLeftPressPixel( QMouseEvent*, Pixel& );

	void mouseMouveOnPixel( Pixel& );

	//	Sur le relachement du boutton de la sourie, envoie le pixel sous
	//	le curseur.
	void mouseLeftReleasePixel( QMouseEvent*, Pixel& );
	
	//	Une modification sur la track a t faite.
	void edited();

private:
	void drawPixel( const QColor& c, const Pixel& pixel );

	int numPixel( const QPoint& cursorPos );

	int	onTrack( const QColor& driverColor, const QColor& traceColor
														,int, int, int, int );
	enum TypeDraw { Rect, Line };
	void draw( TypeDraw, const QColor&,
									const Pixel& pixel0, const Pixel& pixel1 );

	void stateChanged();

	override void paintEvent( QPaintEvent* );
	override void mousePressEvent( QMouseEvent* );
	override void mouseMoveEvent( QMouseEvent* );
    override void mouseReleaseEvent( QMouseEvent* );

	//	Colorise l'ensemble des pixels dans la zone du widget dfinie
	//	par les coordonnes 'coord**'
	//	Renvoie une image de la zone de la piste avant modification.
	TrackPixmap* repaintArea( const ShapePixel&, const Pixel&, const Pixel& );
						

private:
	State		state_;

	int			widthPen_;

	const QSize	minimumSize_;
	Pixel		vpixels_[Dim::maxPixels];	//	Vecteur de tous les pixels
											//	constituant la track.

	//	Mmorise les coordonnes de la ligne trac, la ligne n'altre pas
	//	la piste.
	QRect		selectArea_;

	int			numPixel_;	// mmorise le dernier pixel mit par les
							// fonction mouse[Press|Move|Release]Event

	const ShapePixel&	defaultShape_;
};

inline void WidgetTrack::cancelCleanSelect()
{
	selectArea_.setRect( -1, -1, -1, -1 ); 
}

inline const Pixel&	WidgetTrack::pixel( int num ) const
{
	assert( Dim::pixelValid( num ) );
	return vpixels_[num];
}

inline TrackPixmap*
WidgetTrack::fill( const QColor& c, int numPixel0, int numPixel1 )
{
	assert( Dim::pixelValid( numPixel0 ) && Dim::pixelValid( numPixel1 ) );

	return fill( c, &vpixels_[numPixel0], &vpixels_[numPixel1] );
}

inline void
WidgetTrack::shadowPixel( const QColor& c, int numPixel )
{
	assert( Dim::pixelValid( numPixel ) );
	drawPixel( c, vpixels_[numPixel] );
}

inline void
WidgetTrack::join( const QColor& c, const Pixel& ori, const Pixel& dest )
			
{
	draw( Line, c, ori, dest );
}

inline void
WidgetTrack::select( const QColor& c, const Pixel& ori, const Pixel& dest )
{
	draw( Rect, c, ori, dest );
}

inline void
WidgetTrack::select( const QColor& c, int numPixel0, int numPixel1 )
{
	assert( Dim::pixelValid( numPixel0 ) && Dim::pixelValid( numPixel1 ) );

	select( c, vpixels_[numPixel0], vpixels_[numPixel1] );
}

inline void WidgetTrack::stateChanged()
{
	if ( state_ != Changed )
	{
		state_ = Changed;
		emit edited();
	}
}

inline void WidgetTrack::paintPixel( const QColor& c, int num )
{
	assert( Dim::pixelValid( num ) );
	paintPixel( c, vpixels_[num] );
}

inline int	WidgetTrack::moveCar( const QColor& c, const QColor& t, 
											const Pixel& of, const Pixel& to )
{
	const int ont = onTrack( c, t, of.x(), of.y(), to.x(), to.y() );
	paintPixel( c, to.num() );
	return ont;
}

#endif	//WIDGETTRACK_H


